@@ -1,11 +1,11 @@
+import contextlib
 import json
 from datetime import date
 from datetime import datetime
+from datetime import timedelta
 from time import mktime
 from typing import Iterable
-from typing import List
 from typing import Optional
-from typing import Tuple
 from typing import Union
 
 import click
@@ -13,7 +13,6 @@ import humanize
 import parsedatetime
 import pytz
 from dateutil.tz import tzlocal
-from tabulate import tabulate
 
 from todoman.model import Todo
 from todoman.model import TodoList
@@ -63,94 +62,85 @@ class DefaultFormatter:
         return self.compact_multiple([todo])
 
     def compact_multiple(self, todos: Iterable[Todo], hide_list=False) -> str:
+        # TODO: format lines fuidly and drop the table
+        # it can end up being more readable when too many columns are empty.
+        # show dates that are in the future in yellow (in 24hs) or grey (future)
         table = []
         for todo in todos:
             completed = "X" if todo.is_completed else " "
             percent = todo.percent_complete or ""
             if percent:
                 percent = f" ({percent}%)"
-            priority = self.format_priority_compact(todo.priority)
 
-            due = self.format_datetime(todo.due)
-            now = self.now if isinstance(todo.due, datetime) else self.now.date()
-            if todo.due and todo.due <= now and not todo.is_completed:
-                due = click.style(str(due), fg="red")
+            if todo.categories:
+                categories = " [" + ", ".join(todo.categories) + "]"
+            else:
+                categories = ""
+
+            priority = click.style(
+                self.format_priority_compact(todo.priority),
+                fg="magenta",
+            )
+
+            due = self.format_datetime(todo.due) or "(no due date)"
+            due_colour = self._due_colour(todo)
+            if due_colour:
+                due = click.style(str(due), fg=due_colour)
 
             recurring = "⟳" if todo.is_recurring else ""
 
             if hide_list:
-                summary = "{} {}".format(
-                    todo.summary,
-                    percent,
-                )
+                summary = f"{todo.summary} {percent}"
             else:
                 if not todo.list:
                     raise ValueError("Cannot format todo without a list")
 
-                summary = "{} {}{}".format(
-                    todo.summary,
-                    self.format_database(todo.list),
-                    percent,
-                )
+                summary = f"{todo.summary} {self.format_database(todo.list)}{percent}"
 
+            # TODO: add spaces on the left based on max todos"
+
+            # FIXME: double space when no priority
+            # split into parts to satisfy linter line too long
             table.append(
-                [
-                    todo.id,
-                    f"[{completed}]",
-                    priority,
-                    f"{due} {recurring}",
-                    summary,
-                ]
+                f"[{completed}] {todo.id} {priority} {due} "
+                f"{recurring}{summary}{categories}"
             )
 
-        return tabulate(table, tablefmt="plain")
+        return "\n".join(table)
 
-    def _columnize_text(
-        self,
-        label: str,
-        text: Optional[str],
-    ) -> List[Tuple[Optional[str], str]]:
-        """Display text, split text by line-endings, on multiple colums.
+    def _due_colour(self, todo: Todo) -> str:
+        now = self.now if isinstance(todo.due, datetime) else self.now.date()
+        if todo.due:
+            if todo.due <= now and not todo.is_completed:
+                return "red"
+            elif todo.due >= now + timedelta(hours=24):
+                return "white"
+            elif todo.due >= now:
+                return "yellow"
 
-        Do nothing if text is empty or None.
-        """
-        lines = text.splitlines() if text else None
+        return "white"
 
-        return self._columnize_list(label, lines)
+    def _format_multiline(self, title: str, value: str) -> str:
+        formatted_title = click.style(title, fg="white")
 
-    def _columnize_list(
-        self,
-        label: str,
-        lst: Optional[List[str]],
-    ) -> List[Tuple[Optional[str], str]]:
-        """Display list on multiple columns.
-
-        Do nothing if list is empty or None.
-        """
-
-        rows: List[Tuple[Optional[str], str]] = []
-
-        if lst:
-            rows.append((label, lst[0]))
-            for line in lst[1:]:
-                rows.append((None, line))
-
-        return rows
+        if value.strip().count("\n") == 0:
+            return f"\n\n{formatted_title}: {value}"
+        else:
+            return f"\n\n{formatted_title}:\n{value}"
 
     def detailed(self, todo: Todo) -> str:
         """Returns a detailed representation of a task.
 
         :param todo: The todo component.
         """
-        extra_rows = []
-        extra_rows += self._columnize_text("Description", todo.description)
-        extra_rows += self._columnize_text("Location", todo.location)
+        extra_lines = []
+        if todo.description:
+            extra_lines.append(self._format_multiline("Description", todo.description))
 
-        if extra_rows:
-            return "{}\n\n{}".format(
-                self.compact(todo), tabulate(extra_rows, tablefmt="plain")
-            )
-        return self.compact(todo)
+        if todo.location:
+            extra_lines.append(self._format_multiline("Location", todo.location))
+
+        return f"{self.compact(todo)}{''.join(extra_lines)}"
 
     def format_datetime(self, dt: Optional[date]) -> Union[str, int, None]:
         if not dt:
@@ -160,6 +150,14 @@ class DefaultFormatter:
         elif isinstance(dt, date):
             return dt.strftime(self.date_format)
 
+    def format_categories(self, categories):
+        return ", ".join(categories)
+
+    def parse_categories(self, categories):
+        # existing code assumes categories is list,
+        # but click passes tuple
+        return list(categories)
+
     def parse_priority(self, priority: Optional[str]) -> Optional[int]:
         if priority is None or priority == "":
             return None
@@ -184,6 +182,8 @@ class DefaultFormatter:
         elif 6 <= priority <= 9:
             return "low"
 
+        raise ValueError("priority is an invalid value")
+
     def format_priority_compact(self, priority: Optional[int]) -> str:
         if not priority:
             return ""
@@ -194,6 +194,8 @@ class DefaultFormatter:
         elif 6 <= priority <= 9:
             return "!"
 
+        raise ValueError("priority is an invalid value")
+
     def parse_datetime(self, dt: str) -> Optional[date]:
         if not dt:
             return None
@@ -203,22 +205,16 @@ class DefaultFormatter:
 
     def _parse_datetime_naive(self, dt: str) -> date:
         """Parse dt and returns a naive datetime or a date"""
-        try:
+        with contextlib.suppress(ValueError):
             return datetime.strptime(dt, self.datetime_format)
-        except ValueError:
-            pass
 
-        try:
+        with contextlib.suppress(ValueError):
             return datetime.strptime(dt, self.date_format).date()
-        except ValueError:
-            pass
 
-        try:
+        with contextlib.suppress(ValueError):
             return datetime.combine(
                 self.now.date(), datetime.strptime(dt, self.time_format).time()
             )
-        except ValueError:
-            pass
 
         rv, pd_ctx = self._parsedatetime_calendar.parse(dt)
         if not pd_ctx.hasDateOrTime:
@@ -250,11 +246,13 @@ class PorcelainFormatter(DefaultFormatter):
     def _todo_as_dict(self, todo):
         return {
             "completed": todo.is_completed,
+            "start": self.format_datetime(todo.start),
             "due": self.format_datetime(todo.due),
             "id": todo.id,
             "list": todo.list.name,
             "percent": todo.percent_complete,
             "summary": todo.summary,
+            "categories": todo.categories,
             "priority": todo.priority,
             "location": todo.location,
             "description": todo.description,
@@ -280,7 +278,7 @@ class PorcelainFormatter(DefaultFormatter):
             else:
                 raise ValueError("Priority has to be in the range 0-9")
         except ValueError as e:
-            raise click.BadParameter(str(e))
+            raise click.BadParameter(str(e)) from None
 
     def detailed(self, todo: Todo) -> str:
         return self.compact(todo)
