@@ -8,6 +8,7 @@ from datetime import date
 from datetime import datetime
 from datetime import time
 from datetime import timedelta
+from functools import cached_property
 from os.path import normpath
 from os.path import split
 from typing import Iterable
@@ -29,24 +30,6 @@ logger = logging.getLogger(name=__name__)
 LOCAL_TIMEZONE = tzlocal()
 
 
-class cached_property:
-    """A read-only @property that is only evaluated once. Only usable on class
-    instances' methods.
-    """
-
-    def __init__(self, fget, doc=None):
-        self.__name__ = fget.__name__
-        self.__module__ = fget.__module__
-        self.__doc__ = doc or fget.__doc__
-        self.fget = fget
-
-    def __get__(self, obj, cls):
-        if obj is None:
-            return self
-        obj.__dict__[self.__name__] = result = self.fget(obj)
-        return result
-
-
 class Todo:
     """
     Represents a task/todo, and wrapps around icalendar.Todo.
@@ -71,10 +54,10 @@ class Todo:
 
     def __init__(
         self,
-        filename: str = None,
-        mtime: int = None,
+        filename: str | None = None,
+        mtime: int | None = None,
         new: bool = False,
-        list: TodoList = None,
+        list: TodoList | None = None,
     ):
         """
         Creates a new todo using `todo` as a source.
@@ -297,7 +280,6 @@ class VtodoWriter:
         "due": "due",
         "location": "location",
         "percent_complete": "percent-complete",
-        "priority": "priority",
         "status": "status",
         "created_at": "created",
         "last_modified": "last-modified",
@@ -390,8 +372,8 @@ class VtodoWriter:
                 if component.get("uid", None) == self.todo.uid:
                     cal.subcomponents[index] = vtodo
 
-        with AtomicWriter(path, overwrite=True).open() as f:
-            f.write(cal.to_ical().decode("UTF-8"))
+        with AtomicWriter(path, "wb", overwrite=True).open() as f:
+            f.write(cal.to_ical())
 
     def _write_new(self, path):
         vtodo = self.serialize()
@@ -399,10 +381,10 @@ class VtodoWriter:
         c = icalendar.Calendar()
         c.add_component(vtodo)
 
-        with AtomicWriter(path).open() as f:
+        with AtomicWriter(path, "wb").open() as f:
             c.add("prodid", "io.barrera.todoman")
             c.add("version", "2.0")
-            f.write(c.to_ical().decode("UTF-8"))
+            f.write(c.to_ical())
 
         return vtodo
 
@@ -420,7 +402,7 @@ class Cache:
     may be used for filtering/sorting.
     """
 
-    SCHEMA_VERSION = 7
+    SCHEMA_VERSION = 9
 
     def __init__(self, path: str):
         self.cache_path = str(path)
@@ -445,17 +427,21 @@ class Cache:
         except sqlite3.OperationalError:
             return False
 
+    def drop_tables(self):
+        self._conn.executescript(
+            """
+            DROP TABLE IF EXISTS todos;
+            DROP TABLE IF EXISTS lists;
+            DROP TABLE IF EXISTS files;
+            DROP TABLE IF EXISTS categories;
+        """
+        )
+
     def create_tables(self):
         if self.is_latest_version():
             return
 
-        self._conn.executescript(
-            """
-            DROP TABLE IF EXISTS lists;
-            DROP TABLE IF EXISTS files;
-            DROP TABLE IF EXISTS todos;
-        """
-        )
+        self.drop_tables()
 
         self._conn.execute('CREATE TABLE IF NOT EXISTS meta ("version" INT)')
 
@@ -490,6 +476,19 @@ class Cache:
         """
         )
 
+        self._conn.execute(
+            """
+            CREATE TABLE IF NOT EXISTS categories (
+                "id" INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
+                "todos_id" INTEGER NOT NULL,
+                "category" TEXT,
+
+                CONSTRAINT category_unique UNIQUE (todos_id,category),
+                FOREIGN KEY(todos_id) REFERENCES todos(id) ON DELETE CASCADE
+            );
+        """
+        )
+
         self._conn.execute(
             """
             CREATE TABLE IF NOT EXISTS todos (
@@ -510,7 +509,6 @@ class Cache:
                 "status" TEXT,
                 "description" TEXT,
                 "location" TEXT,
-                "categories" TEXT,
                 "sequence" INTEGER,
                 "last_modified" INTEGER,
                 "rrule" TEXT,
@@ -558,7 +556,7 @@ class Cache:
                 ),
             )
         except sqlite3.IntegrityError as e:
-            raise exceptions.AlreadyExists("list", name) from e
+            raise exceptions.AlreadyExistsError("list", name) from e
 
         return self.add_list(name, path, colour, mtime)
 
@@ -579,7 +577,21 @@ class Cache:
                 ),
             )
         except sqlite3.IntegrityError as e:
-            raise exceptions.AlreadyExists("file", list_name) from e
+            raise exceptions.AlreadyExistsError("file", list_name) from e
+
+    def add_category(self, todos_id, category):
+        try:
+            self._conn.execute(
+                """
+               INSERT INTO categories (
+                   todos_id,
+                   category
+               ) VALUES (?, ?);
+               """,
+                (todos_id, category),
+            )
+        except sqlite3.IntegrityError as e:
+            raise exceptions.AlreadyExistsError("category", category) from e
 
     def _serialize_datetime(
         self,
@@ -613,13 +625,6 @@ class Cache:
 
         return rrule.to_ical().decode()
 
-    def _serialize_categories(self, todo, field) -> str:
-        categories = todo.get(field, [])
-        if not categories:
-            return ""
-
-        return ",".join([str(category) for category in categories.cats])
-
     def add_vtodo(self, todo: icalendar.Todo, file_path: str, id=None) -> int:
         """
         Adds a todo into the cache.
@@ -645,11 +650,10 @@ class Cache:
                 status,
                 description,
                 location,
-                categories,
                 sequence,
                 last_modified,
                 rrule
-            ) VALUES ({}?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?,
+            ) VALUES ({}?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?,
                 ?)
             """
 
@@ -675,7 +679,6 @@ class Cache:
             todo.get("status", "NEEDS-ACTION"),
             todo.get("description", None),
             todo.get("location", None),
-            self._serialize_categories(todo, "categories"),
             todo.get("sequence", 1),
             self._serialize_datetime(todo, "last-modified")[0],
             self._serialize_rrule(todo, "rrule"),
@@ -691,17 +694,22 @@ class Cache:
         try:
             cursor.execute(sql, params)
             rv = cursor.lastrowid
+            assert rv is not None
         finally:
             cursor.close()
 
+        if todo.get("categories"):
+            for category in todo.get("categories").cats:
+                self.add_category(rv, category)
+
         return rv
 
     def todos(
         self,
         lists=(),
+        categories=None,
         priority=None,
         location="",
-        category="",
         grep="",
         sort=(),
         reverse=True,
@@ -723,7 +731,7 @@ class Cache:
         :param list lists: Only return todos for these lists.
         :param str location: Only return todos with a location containing this
             string.
-        :param str category: Only return todos with a category containing this
+        :param str categories: Only return todos with a category containing this
             string.
         :param str grep: Filter common fields with this substring.
         :param list sort: Order returned todos by these fields. Field names
@@ -758,15 +766,16 @@ class Cache:
             q = ", ".join(["?"] * len(lists))
             extra_where.append(f"AND files.list_name IN ({q})")
             params.extend(lists)
+        if categories:
+            category_slots = ", ".join(["?"] * len(categories))
+            extra_where.append(f"AND upper(categories.category) IN ({category_slots})")
+            params = params + [category.upper() for category in categories]
         if priority:
             extra_where.append("AND PRIORITY > 0 AND PRIORITY <= ?")
             params.append(f"{priority}")
         if location:
             extra_where.append("AND location LIKE ?")
             params.append(f"%{location}%")
-        if category:
-            extra_where.append("AND categories LIKE ?")
-            params.append(f"%{category}%")
         if grep:
             # # requires sqlite with pcre, which won't be available everywhere:
             # extra_where.append('AND summary REGEXP ?')
@@ -811,10 +820,14 @@ class Cache:
             order = order.replace(" DESC", " asc").replace(" ASC", " desc")
 
         query = """
-              SELECT todos.*, files.list_name, files.path
-                FROM todos, files
-               WHERE todos.file_path = files.path {}
-            ORDER BY {}
+        SELECT DISTINCT todos.*, files.list_name, files.path,
+          group_concat(category) AS categories
+        FROM todos, files
+        LEFT JOIN categories
+        ON categories.todos_id = todos.id
+        WHERE todos.file_path = files.path {}
+        GROUP BY uid
+        ORDER BY {}
         """.format(
             " ".join(extra_where),
             order,
@@ -856,6 +869,12 @@ class Cache:
         else:
             return datetime.fromtimestamp(dt, LOCAL_TIMEZONE)
 
+    def _categories_from_db(self, categories):
+        if categories:
+            return categories.split(",")
+
+        return []
+
     def _todo_from_db(self, row: dict) -> Todo:
         todo = Todo()
         todo.id = row["id"]
@@ -863,6 +882,7 @@ class Cache:
         todo.summary = row["summary"]
         todo.due = self._date_from_db(row["due"], row["due_dt"])
         todo.start = self._date_from_db(row["start"], row["start_dt"])
+        todo.categories = self._categories_from_db(row["categories"])
         todo.priority = row["priority"]
         todo.created_at = self._datetime_from_db(row["created_at"])
         todo.completed_at = self._datetime_from_db(row["completed_at"])
@@ -871,6 +891,8 @@ class Cache:
         todo.status = row["status"]
         todo.description = row["description"]
         todo.location = row["location"]
+
+        logger.debug("todo.categories: %s\n", todo.categories)
         todo.sequence = row["sequence"]
         todo.last_modified = row["last_modified"]
         todo.list = self.lists_map[row["list_name"]]
@@ -908,16 +930,20 @@ class Cache:
         # XXX: DON'T USE READ_ONLY
         result = self._conn.execute(
             """
-            SELECT todos.*, files.list_name, files.path
-              FROM todos, files
+            SELECT todos.*, files.list_name, files.path,
+              group_concat(category) AS categories
+            FROM todos, files
+            LEFT JOIN categories
+            ON categories.todos_id = todos.id
             WHERE files.path = todos.file_path
               AND todos.id = ?
+            GROUP BY uid
         """,
             (id,),
         ).fetchone()
 
         if not result:
-            raise exceptions.NoSuchTodo(id)
+            raise exceptions.NoSuchTodoError(id)
 
         if not read_only:
             count = self._conn.execute(
@@ -930,7 +956,7 @@ class Cache:
                 (result["path"],),
             ).fetchone()
             if count["c"] > 1:
-                raise exceptions.ReadOnlyTodo(result["path"])
+                raise exceptions.ReadOnlyTodoError(result["path"])
 
         return self._todo_from_db(result)
 
@@ -947,7 +973,7 @@ class Cache:
 
 
 class TodoList:
-    def __init__(self, name: str, path: str, colour: str = None):
+    def __init__(self, name: str, path: str, colour: str | None = None):
         self.path = path
         self.name = name
         self.colour = colour
@@ -960,6 +986,8 @@ class TodoList:
         except OSError:
             logger.debug("No colour for list %s", path)
 
+        return None
+
     @staticmethod
     def name_for_path(path: str) -> str:
         try:
@@ -1036,7 +1064,7 @@ class Database:
 
             try:
                 self.cache.add_file(list_name, entry_path, mtime)
-            except exceptions.AlreadyExists:
+            except exceptions.AlreadyExistsError:
                 logger.debug("File already in cache: %s", entry_path)
                 continue
 
@@ -1102,5 +1130,4 @@ class Database:
 
 
 def _getmtime(path: str) -> int:
-    stat = os.stat(path)
-    return getattr(stat, "st_mtime_ns", stat.st_mtime)
+    return os.stat(path).st_mtime_ns
